package at.jku.fim.datalinksimulation.genericdevices;

import java.util.HashMap;

import at.jku.fim.datalinksimulation.layer1.data.RawData;
import at.jku.fim.datalinksimulation.layer1.wire.Wire;
import at.jku.fim.datalinksimulation.layer1.wire.WireConnector;
import at.jku.fim.datalinksimulation.layer1.wire.WiringException;
import at.jku.fim.datalinksimulation.layer2.filters.EasyL2CRCCheckFilter;
import at.jku.fim.datalinksimulation.layer2.filters.EasyL2DecodingFilter;
import at.jku.fim.datalinksimulation.layer2.filters.EasyL2DestinationFilter;
import at.jku.fim.datalinksimulation.layer2.filters.EasyL2EncodingFilter;
import at.jku.fim.datalinksimulation.layer2.filters.EasyL2SanityCheckFilter;
import at.jku.fim.datalinksimulation.layer2.filters.EasyL2v2DestuffingFilter;
import at.jku.fim.datalinksimulation.layer2.filters.EasyL2v2StuffingFilter;
import at.jku.fim.datalinksimulation.layer7.Application;
import at.jku.fim.datalinksimulation.networkstack.NetworkStack;
import at.jku.fim.datalinksimulation.networkstack.NetworkStackParameters;
import at.jku.fim.datalinksimulation.networkstack.incoming.DecodingNetworkFilter;
import at.jku.fim.datalinksimulation.networkstack.incoming.IncomingNetworkStackConfiguration;
import at.jku.fim.datalinksimulation.networkstack.outgoing.EncodingNetworkFilter;
import at.jku.fim.datalinksimulation.networkstack.outgoing.OutgoingNetworkStackConfiguration;
import at.jku.fim.datalinksimulation.networkstack.outgoing.OutgoingNetworkStackObserver;
import at.jku.fim.datalinksimulation.utils.ConversionUtils;
import at.jku.fim.datalinksimulation.utils.Logger;


/**
 * <p>Title: OSI Simulation Framework</p>
 * A <code>NetworkStation</code> represents a system, which can run
 * different applications. Each of these have access to a single network 
 * stack for performing communication.
 * <p>Copyright: (c) 2004-05</p>
 * <p>Company: FIM@JKU, www.fim.uni-linz.ac.at</p>
 * @author Andreas Putzinger
 * @version $Id: NetworkStation.java,v 1.2 2005/12/13 15:24:03 msonntag Exp $
 */
public class NetworkStation implements WireConnector {

    /** Internal network stack reference */
    protected NetworkStack networkStack;

    /**
     * Running application get a Process ID (pid). The
     * mapping between the actual application and the
     * PID is stored in this HashMap
     */
    protected HashMap PIDApplicationMapping = new HashMap();

    /** A PID counter so that a PID is unique. */
    protected int PID = 0;

    protected Logger logger;

    /**
     * Constructor; Each network station in a system should have a unique ID
     * @param stationID Unique station ID
     */
    public NetworkStation(Object stationID) {
        logger = new Logger("Network station: " + stationID.toString());
        networkStack = new NetworkStack(new IncomingNetworkStackConfiguration() {

			    protected DecodingNetworkFilter[][] filters = new DecodingNetworkFilter[6][];

			    {
			        filters[1] = new DecodingNetworkFilter[5];
			        filters[1][0] = new EasyL2v2DestuffingFilter();
			        filters[1][1] = new EasyL2CRCCheckFilter();
			        filters[1][2] = new EasyL2DecodingFilter();
			        filters[1][3] = new EasyL2SanityCheckFilter();
			        filters[1][4] = new EasyL2DestinationFilter();
			    }

			    /*
			     *  (non-Javadoc)
			     * @see at.jku.fim.datalinksimulation.networkstack.IncomingNetworkStackConfiguration#getDecodingFilters(int)
			     */
			    public DecodingNetworkFilter[] getDecodingFilters(int osiLevel) {
			        if (osiLevel >= 1 && osiLevel <= 6) {
			            return filters[osiLevel - 1];
			        } else
			            return null;

			    }
			}, new OutgoingNetworkStackConfiguration () {

	    	    protected EncodingNetworkFilter[][] filters = new EncodingNetworkFilter[6][];

	    	    {
	    	        filters[1] = new EncodingNetworkFilter[2];
	    	        filters[1][0] = new EasyL2EncodingFilter();
	    	        filters[1][1] = new EasyL2v2StuffingFilter();
	    	    }

	    	    /*
	    	     *  (non-Javadoc)
	    	     * @see at.jku.fim.datalinksimulation.networkstack.OutgoingNetworkStackConfiguration#getEncodingFilters(int)
	    	     */
	    	    public EncodingNetworkFilter[] getEncodingFilters(int osiLevel) {
	    	        if (osiLevel >= 1 && osiLevel <= 6) {
	    	            return filters[osiLevel - 1];
	    	        } else
	    	            return null;
	    	    }
	    	}, stationID);
        
        networkStack.setParentLogger(logger);
        
    }

    /**
     * @return Returns the stationID.
     */
    public Object getStationID() {
        return networkStack.getStationID();
    }

    /**
     * @param stationID The stationID to set.
     */
    public void setStationID(Object stationID) {
        networkStack.setStationID(stationID);
    }

    /**
     * Starts an application on this network station
     * @param applicationObj
     * @return Process ID
     */
    public int runApplication(Application applicationObj) {
        int pid = -1;
        if (applicationObj != null) {
            networkStack.subscribeApplication(applicationObj);
            applicationObj.setNetworkStack(networkStack);
            pid = PID++;
            PIDApplicationMapping.put(new Integer(pid), applicationObj);
            if (applicationObj.getLogger() != null)
            	logger.addChildLogger(applicationObj.getLogger());
        }
        if(applicationObj instanceof Runnable) {
        	(new Thread((Runnable)applicationObj)).start();
        }
        return pid;
    }

    /**
     * 
     * @param applicationPID
     */
    public void removeApplication(int applicationPID) {
        Object obj = PIDApplicationMapping.get(new Integer(applicationPID));
        if (obj != null) {
            Application applicationObj = (Application) obj;
            networkStack.unsubscribeApplication(applicationObj);
            PIDApplicationMapping.remove(new Integer(applicationPID));
        }
    }

    /**
     * informAboutWireData
     *
     * @param sender Wire
     * @param data RawData
     */
    public void informAboutDataOnWire(Wire sender, RawData data) {
        logger.log("Received data");
        logger.log("  Plaintext: <" + data.stringRepresentation() + ">");
        logger.log("  Hex: <" + ConversionUtils.stringToHex(data.toString()) + ">");
        logger
                .log("Forwarding data to my network stack for further processing.");
        networkStack.receiveFromWorld(data, 0);
    }

    public NetworkStack getNetworkStack() {
        return networkStack;
    }

    /*
	 * (non-Javadoc)
	 * 
	 * @see at.jku.fim.datalinksimulation.layer1.wire.WireConnector#bePlugged(at.jku.fim.datalinksimulation.layer1.wire.Wire,
	 *      int)
	 */
	public void bePlugged(final Wire wire, int portNr) throws WiringException {
        networkStack.connectPhysical(new OutgoingNetworkStackObserver() {
			public void informAboutNewData(RawData data, NetworkStackParameters params) {
				wire.send(NetworkStation.this, data);
			}
        });
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see at.jku.fim.datalinksimulation.layer1.wire.WireConnector#beUnplugged(at.jku.fim.datalinksimulation.layer1.wire.Wire,
	 *      int)
	 */
	public void beUnplugged(final Wire wire, int portNr) throws WiringException {
		// not yet implemented
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see at.jku.fim.datalinksimulation.layer1.wire.WireConnector#getNumberOfTotalPorts()
	 */
	public int getNumberOfTotalPorts() {
		return 1;
	}
}