package at.jku.fim.datalinksimulation.networkstack.incoming;

import java.util.Iterator;
import java.util.Vector;

import at.jku.fim.datalinksimulation.layer1.data.RawData;
import at.jku.fim.datalinksimulation.networkstack.InvalidFilterConfigurationException;
import at.jku.fim.datalinksimulation.networkstack.NetworkStackParameters;
import at.jku.fim.datalinksimulation.utils.ConversionUtils;
import at.jku.fim.datalinksimulation.utils.Logger;

/**
 * <p>Title: OSI Simulation Framework</p>
 * <p>Copyright: (c) 2004-05</p>
 * <p>Company: FIM@JKU, www.fim.uni-linz.ac.at</p>
 * @author Andreas Putzinger
 * @version $Id: IncomingNetworkStack.java,v 1.1 2005/10/27 13:10:11 aputzinger Exp $
 */

public class IncomingNetworkStack {

    /**
     * network stack parameter name for the ID of the host station
     */
    public static String NETWORK_PARAM_STATION_ID = "THIS_STATION";

    /**
     * Host station ID
     */
    protected Object stationId;

    /**
     * Filter configuration for incoming network stack
     */
    protected IncomingNetworkStackConfiguration configuration = null;

    /**
     * Registered interested objects
     */
    protected Vector subscribers = new Vector();

    /**
     * Associated logger instance
     */
    protected Logger logger;

    /**
     * Constrcutor
     * @param configuration Filter configuration
     * @param stationID Station ID
     */
    public IncomingNetworkStack(
            IncomingNetworkStackConfiguration configuration, Object stationID) {
        try {
            setConfiguration(configuration);
        } catch (InvalidFilterConfigurationException e) {
            e.printStackTrace();
        }
        this.stationId = stationID;
        logger = new Logger("Incoming network stack for station "
                + stationID.toString());
    }

    /**
     * Subscribe to results of the incoming network stack; The only one who currently subscribes it is the NetworkStack class
     * @param newObserver
     */
    public void subscribe(IncomingNetworkStackObserver newObserver) {
        subscribers.add(newObserver);
    }

    /**
     * Unsubscrive
     * @param registeredObserver
     */
    public void unsubscribe(IncomingNetworkStackObserver registeredObserver) {
        subscribers.remove(registeredObserver);
    }

    /**
     * Check if instance is subsribed
     * @param observer
     * @return Returns, if instance if subscribed or not
     */
    boolean isSubscribed(Object observer) {
        return subscribers.contains(observer);
    }

    /**
     * Set a new filter configuration
     * @param configuration Filter configuration
     */
    public void setConfiguration(IncomingNetworkStackConfiguration configuration) throws InvalidFilterConfigurationException{
        if (configuration == null) {
            this.configuration = null;
        } else {
            for (int osiLevel = 1; osiLevel <= 6; osiLevel++) {
                DecodingNetworkFilter[] filters = configuration
                        .getDecodingFilters(osiLevel);
                if (filters != null) {
                    for (int filterNum = 0; filterNum < filters.length; filterNum++) {
                        DecodingNetworkFilter filter = filters[filterNum];
                        int[] supportedFilterOSILayers = filter
                                .getFilterOSILayer();
                        boolean supported = false;
                        if (supportedFilterOSILayers != null) {
                            for (int i = 0; i < supportedFilterOSILayers.length; i++) {
                                if (supportedFilterOSILayers[i] == osiLevel) {
                                    supported = true;
                                    break;
                                }
                            }
                        }
                        if (!supported) {
                            throw new InvalidFilterConfigurationException("Filter(s) does not support usage in specified ISO/OSI layer.");
                        }
                    }
                }
            }
            // Everything is OK, set configuration active
            this.configuration = configuration;
        }
    }

    /**
     * @return Returns current Filter configuration
     */
    public IncomingNetworkStackConfiguration getConfiguration() {
        return configuration;
    }

    /**
     * This method is called whenever data is received on the connected wire
     * @param params network stack params
     * @param data received data
     */
    protected void receive(NetworkStackParameters params, RawData data) {
        logger.log("Receiving following data:");
        logger.log("    Plaintext: <" + data.stringRepresentation() + ">");
        logger.log("    Hex: <" + ConversionUtils.stringToHex(data.toString()) + ">\n");
        logger.log("Searching for filters to apply on data");

        if (configuration != null) {
            for (int osiLayer = 1; osiLayer <= 6; osiLayer++) {
                DecodingNetworkFilter[] filters = configuration
                        .getDecodingFilters(osiLayer);
                if (filters != null) {
                    int filterNum = 0;
                    while (filterNum < filters.length && data != null) {
                        DecodingNetworkFilter filter = filters[filterNum++];
                        logger.log("+++ Applying Filter <"
                                + filter.getDescription() + ">");
                        data = filter.decode(params, data, logger);
                    }
                    if (data == null) {
                        logger
                                .log("Last filter consumed package; no further processing; package is ignored; application will never see package");
                        return;
                    }
                }
            }
        }
        logger.log("Resulting Data (all filters are applied): ");
        logger.log("    Plaintext: <" + data.stringRepresentation() + ">");
        logger.log("    Hex: <" + ConversionUtils.stringToHex(data.toString()) + ">");
        Iterator it = subscribers.iterator();
        while (it.hasNext()) {
            IncomingNetworkStackObserver observer = (IncomingNetworkStackObserver) it
                    .next();
            observer.informAboutNewData(params, data);
        }
    }

    /*
	 * (non-Javadoc)
	 * 
	 * @see at.jku.fim.datalinksimulation.layer1.wire.WireObserver#informAboutWireData(at.jku.fim.datalinksimulation.data.RawData,
	 *      java.util.Vector)
	 */
    public void receiveNewData(RawData data, NetworkStackParameters params) {
    	if (params == null)
    		params = new NetworkStackParameters();
    	
        params.setParameter(NetworkStackParameters.GLOBAL_PARAMETER,
                NETWORK_PARAM_STATION_ID, this.stationId);
        receive(params, data);
    }

	/**
	 * @return Returns the stationId.
	 */
	public Object getStationId() {
		return stationId;
	}

	/**
	 * @param stationId The stationId to set.
	 */
	public void setStationId(Object stationId) {
		this.stationId = stationId;
	}

	/**
	 * @param parent Parent logger
	 */
	public void setParentLogger(Logger parent) {
		parent.addChildLogger(logger);
	}

}