package at.jku.fim.datalinksimulation.networkstack.outgoing;

import java.util.Iterator;
import java.util.Vector;

import at.jku.fim.datalinksimulation.layer1.data.RawData;
import at.jku.fim.datalinksimulation.networkstack.InvalidFilterConfigurationException;
import at.jku.fim.datalinksimulation.networkstack.NetworkStackParameters;
import at.jku.fim.datalinksimulation.utils.ConversionUtils;
import at.jku.fim.datalinksimulation.utils.Logger;

/**
 * <p>Title: OSI Simulation Framework</p>
 * <p>Copyright: (c) 2004-05</p>
 * <p>Company: FIM@JKU, www.fim.uni-linz.ac.at</p>
 * @author Andreas Putzinger
 * @version $Id: OutgoingNetworkStack.java,v 1.1 2005/10/27 13:10:11 aputzinger Exp $
 */

public class OutgoingNetworkStack {

    /**
     * Associated Filter configuration 
     */
    protected OutgoingNetworkStackConfiguration configuration = null;

    /**
     * Connected oberservers
     */
    protected Vector oberservers = new Vector();

    /**
     * Associated logger instance
     */
    protected Logger logger;

    /**
     * Identifier of host
     */
    protected Object stationID;

    /**
     * Constructor
     * @param configuration Filter configuration
     * @param stationID StationID of host
     */
    public OutgoingNetworkStack(
            OutgoingNetworkStackConfiguration configuration, Object stationID) {
        try {
            setConfiguration(configuration);
        } catch (InvalidFilterConfigurationException e) {
            e.printStackTrace();
        }
        this.stationID = stationID;
        logger = new Logger("Outgoing network stack for station "
                + stationID.toString());
    }

    /**
     * Connect this instance to a wire
     * @param newObserver 
     */
    public void subscribe(OutgoingNetworkStackObserver newObserver) {
        oberservers.add(newObserver);
    }

    /**
     * Disconnect wire
     * @param registeredObserver
     */
    public void unsubscribe(OutgoingNetworkStackObserver registeredObserver) {
        oberservers.remove(registeredObserver);
    }

    /**
     * Set a new filter configuration
     * @param configuration
     * @throws InvalidFilterConfigurationException
     */
    public void setConfiguration(OutgoingNetworkStackConfiguration configuration) throws InvalidFilterConfigurationException {
        if (configuration == null) {
            this.configuration = null;
        } else {
            for (int osiLevel = 1; osiLevel <= 6; osiLevel++) {
                EncodingNetworkFilter[] filters = configuration
                        .getEncodingFilters(osiLevel);
                if (filters != null) {
                    for (int filterNum = 0; filterNum < filters.length; filterNum++) {
                        EncodingNetworkFilter filter = filters[filterNum];
                        int[] supportedFilterOSILayers = filter
                                .getFilterOSILayer();
                        boolean supported = false;
                        if (supportedFilterOSILayers != null) {
                            for (int i = 0; i < supportedFilterOSILayers.length; i++) {
                                if (supportedFilterOSILayers[i] == osiLevel) {
                                    supported = true;
                                    break;
                                }
                            }
                        }
                        if (!supported) {
                            throw new InvalidFilterConfigurationException("Filter(s) does not support usage in specified ISO/OSI layer.");
                        }
                    }
                }
            }
            // Everything is OK, set configuration active
            this.configuration = configuration;
        }
    }

    /**
     * @return Returns current filter configuration
     */
    public OutgoingNetworkStackConfiguration getConfiguration() {
        return configuration;
    }

    /**
     * Send something through the network stack to another station
     * @param params network stack params
     * @param data data to send
     */
    public void send(NetworkStackParameters params, RawData data) {
        logger.log("Send following data:");
        logger.log("     Plaintext: <" + data.stringRepresentation() + ">");
        logger.log("     Hex: <" + ConversionUtils.stringToHex(data.toString()) + ">\n");
        logger.log("Searching for filters to apply on data");
        if (configuration != null) {
            for (int osiLayer = 6; osiLayer >= 1; osiLayer--) {
                EncodingNetworkFilter[] filters = configuration
                        .getEncodingFilters(osiLayer);
                if (filters != null) {
                    int filterNum = 0;
                    while (filterNum < filters.length && data != null) {
                        EncodingNetworkFilter filter = filters[filterNum++];
                        logger.log("+++ Applying Filter <"
                                + filter.getDescription() + ">");
                        data = filter.encode(params, data, logger);
                    }
                    if (data == null) {
                        logger
                                .log("Last filter consumed package; no further processing; package is ignored");
                        return;
                    }
                }
            }
        }
        logger.log("Resulting Data (all filters are applied): ");
        logger.log("     Plaintext: <" + data.stringRepresentation() + ">");
        logger.log("     Hex: <" + ConversionUtils.stringToHex(data.toString()) + ">");
        Iterator it = oberservers.iterator();
        while (it.hasNext()) {
        	OutgoingNetworkStackObserver observer = (OutgoingNetworkStackObserver) it.next();
            logger.log("Sending data to wire");
            observer.informAboutNewData(data, params);
        }
    }

	/**
	 * @return Returns the stationID.
	 */
	public Object getStationID() {
		return stationID;
	}

	/**
	 * @param stationID The stationID to set.
	 */
	public void setStationID(Object stationID) {
		this.stationID = stationID;
	}

	/**
	 * @param parent Parent logger
	 */
	public void setParentLogger(Logger parent) {
		parent.addChildLogger(logger);
	}

}